/*      	 
 * Copyright 2002 Slicer/HackersQuest (slicer@hackersquest.org)
 *
 * This file is part of Odin's Eye.
 *
 * Odin's Eye is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Odin's Eye is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

class MapElementNode {
  public int x, y, z;

  MapElementNode(int xv, int yv, int zv) {
    x=xv;
    y=yv;
    z=zv;
  }

  double distLine(double x1, double y1, double z1, double x2, double y2, double z2) {
    double u;
    double ddist;
    double xm, ym, zm;
    double dist; 
    double tdist;    

    ddist=dist3d(x1,y1,z1,x2,y2,z2);

    if (ddist <= 0.0) {
      return dist3d(x1,y1,z1);
    }

    u = (x-x1)*(x2-x1)+(y-y1)*(y2-y1)+(z-z1)*(z2-z1);
    u = u / (ddist*ddist);

    xm = x1 + u * (x2-x1);
    ym = y1 + u * (y2-y1);
    zm = z1 + u * (z2-z1);

    dist=dist3d(xm,ym,zm);
    tdist = dist3d(x1,y1,z1);
    if (tdist < dist)
      dist=tdist;
    tdist = dist3d(x2,y2,z2);
    if (tdist < dist)
      dist=tdist;

    return dist;
  }

  double distLine(MapElementNode a, MapElementNode b) {
    return distLine(a.x, a.y, a.z, b.x, b.y, b.z);
  }

  static double dist3d(double x1, double y1, double z1, double x2, double y2, double z2) {
    double xd,yd,zd;
    xd=x1-x2;
    yd=y1-y2;
    zd=z1-z2;
    return Math.sqrt(xd*xd+yd*yd+zd*zd);
  }

  double dist3d(double x1, double y1, double z1) {
    return dist3d(x,y,z,x1,y1,z1);
  }

  double dist3d(MapElementNode a) {
    return dist3d(a.x,a.y,a.z);
  }
}

public class MapElementLine extends MapElement {
  boolean recalc;
  boolean dirty;

  Rectangle size;
  int zavg;
  boolean filled;
  
  MapElementNode node;
  LinkedList nodes;

  public MapElementLine() {
    super();
    title = "Line";
    node = null;
    nodes = new LinkedList();
    dirty = true;
    recalc = true;
    size = new Rectangle();
  }

  public void add(int x, int y, int z) {
    recalc=dirty=true;

    MapElementNode n=new MapElementNode(x, y, z);
    int i;
    i=nodes.indexOf(node);
    if ((i==-1) || (i==nodes.size()))
      nodes.add(n);
    else
      nodes.add(i, n);
    node=n;
  }

  public void deleteSelected() {
    if (node == null)
      return;

    recalc=dirty=true;

    int i=nodes.indexOf(node);
    nodes.remove(node);
    if (nodes.size() == 0) {
      node=null;
      return;
    }
    if (i>0)
      i--;
    node=(MapElementNode) nodes.get(i);
  }

  public boolean getFilled() {
    return filled;
  }

  public void setFilled(boolean f) {
    filled=f;
  }

  public void gotoFirst() {
    if (nodes.size() == 0)
      return;
    node=(MapElementNode) nodes.get(0);
  }
  
  public void gotoLast() {
    if (nodes.size() == 0)
      return;
    node=(MapElementNode) nodes.get(nodes.size()-1);
  }

  public void flatten(int nz) {
    recalc=dirty=true;

    MapElementNode objs[]=(MapElementNode[]) nodes.toArray(new MapElementNode[0]);
    for(int i=0;i<objs.length;i++)
      objs[i].z=nz;
  }

  public void paint(Graphics2D g, Map m) {
    calc();

    if (nodes.size() == 0)
      return;

    MapElementNode objs[]=(MapElementNode[]) nodes.toArray(new MapElementNode[0]);

    int x[]=new int[objs.length];
    int y[]=new int[objs.length];

    for(int i=0;i<objs.length;i++) {
      x[i]=m.transx(objs[i].x, objs[i].y);
      y[i]=m.transy(objs[i].x, objs[i].y);
    }

    if (filled && objs.length > 2) {
      g.setColor(new Color(col.getRed() / 2, col.getGreen() / 2, col.getBlue() / 2));

      Polygon p=new Polygon();
      for(int i=0;i<objs.length;i++) 
        p.addPoint(x[i],y[i]);
      g.fill(p);
    }

    g.setColor(col);

    if (m.getSelected() == this) {
      for(int i=0;i<objs.length;i++) {
        int w;
        if (objs[i] == node) 
          w=5;
        else if (i==0)
          w=4;
        else
          w=2;
        g.fillOval(x[i]-w,y[i]-w,2*w+1,2*w+1);
      }
    }

    if (objs.length < 1)
      return;

    for(int i=0;i<objs.length-1;i++) {
      g.drawLine(x[i],y[i],x[i+1],y[i+1]);
    }

  }

  public boolean visible(Rectangle bounds) {
    calc();
    return bounds.intersects(size);
  }

  public double clickAt(Point p) {
    double best = -1.0;
    MapElementNode objs[]=(MapElementNode []) nodes.toArray(new MapElementNode[0]);
    double xd, yd, d;

    node = null;

    for(int i=0;i<objs.length;i++) {
      d=p.distance(objs[i].x,objs[i].y);
      if ((best < 0.0) || (d < best)) {
        best=d;
        node=objs[i];
      }
    }
    return best;
  }

  public String toString() {
    ArrayList al=new ArrayList();
    MapElementNode objs[]=(MapElementNode []) nodes.toArray(new MapElementNode[0]);
    if (filled)
      al.add("F");
    else
      al.add("M");
    al.add(title);
    al.add(MapEditor.colorToString(col));
    al.add(String.valueOf(objs.length));
    for (int i=0;i<objs.length;i++) {
      al.add(String.valueOf(objs[i].x));
      al.add(String.valueOf(objs[i].y));
      al.add(String.valueOf(objs[i].z));
    }
    return join(al);
  }

  public boolean fromString(String s[]) {
    if (s.length < 4)
      return false;

    if (! super.fromString(s))
      return false;
    try {
      int n, x, y, z;
      n=Integer.valueOf(s[3]).intValue();
      if (s.length != (n * 3)+4) 
        return false;
      for(int i=0;i<n;i++) {
        x=Math.abs(Integer.valueOf(s[i*3+4]).intValue());
        y=Math.abs(Integer.valueOf(s[i*3+5]).intValue());
        z=Math.abs(Integer.valueOf(s[i*3+6]).intValue());
        add(x,y,z);
      }
    } catch (NumberFormatException e) {
      return false;
    }
    return true;
  }

  public int getZ() {
    calc();
    return zavg;
  }

  public int getNodeX() {
    return node.x;
  }

  public int getNodeY() {
    return node.y;
  }

  public void calc() {
    if (! recalc)
      return;

    recalc = false;

    long ztot;
    MapElementNode objs[]=(MapElementNode []) nodes.toArray(new MapElementNode[0]);
    
    ztot=0;

    size=new Rectangle();

    for(int i=0;i<objs.length;i++) {
      size.add(objs[i].x,objs[i].y);
      ztot+=objs[i].z;
    }

    if (objs.length != 0) {   
      zavg =(int) ( ztot / objs.length);
    } else {
      zavg = 0;
    }
  }

  public boolean cleanAndValidate() {
    if (! dirty) 
      return true;
    if (! super.cleanAndValidate())
      return false;
    if (nodes.size() < 2)
      return false;

    douglasPeucker(2.0);

    if (nodes.size() == 2) {
      MapElementNode a=(MapElementNode) nodes.get(0);
      MapElementNode b=(MapElementNode) nodes.get(1);
      if (a.dist3d(b) < 2.0)
        return false;
    }

    dirty = false;
   
    return true;
  }

  public MapElementNode []douglasPeucker(MapElementNode objs[], double sigma) {
     int besti;
     double bestdist;
     double dist;
     MapElementNode p1, p2;
     MapElementNode p;
     MapElementNode r[];
     
     if (objs.length < 3)
       return objs;
     
     besti = 0;
     bestdist = -1.0;

     p1=objs[0];
     p2=objs[objs.length-1];
     for(int i=1;i<objs.length-1;i++) {
       p=objs[i];
       dist=p.distLine(p1,p2);
       if  (dist > bestdist) {
         bestdist=dist;
         besti=i;
       }
     }

     if (bestdist > sigma) {
       MapElementNode a[], b[];
       a=new MapElementNode[besti+1];
       b=new MapElementNode[objs.length-besti];
       for (int i=0;i<=besti;i++) {
         a[i]=objs[i];
       }
       for (int i=0;i<objs.length-besti; i++) {
         b[i]=objs[besti+i];
       }
       a=douglasPeucker(a, sigma);
       b=douglasPeucker(b, sigma);
       r=new MapElementNode[a.length + b.length - 1];
       for (int i=0;i<a.length;i++) 
         r[i]=a[i];
       for (int i=1;i<b.length;i++)
         r[a.length+i-1]=b[i];
     } else {
       r=new MapElementNode[2];
       r[0]=p1;
       r[1]=p2;
     }
     return r;
  }

  public void douglasPeucker(double sigma) {
    if (nodes.size() < 3)
      return;
    MapElementNode objs[]=(MapElementNode []) nodes.toArray(new MapElementNode[0]);
    objs=douglasPeucker(objs, sigma);

    int onodes=nodes.size();

    nodes=new LinkedList();
    for (int i=0;i<objs.length;i++)
      nodes.add(objs[i]);

    int nnodes=nodes.size();

    if (onodes != nnodes) {
      System.out.println("Line "+title+" had "+(onodes-nnodes)+" redundant nodes removed");
    }

    recalc=dirty=true;
  }

}
